module Ubc.Parse.Syntax.Struct
( Struct(..)
, parse
)
where

import Control.Monad ((<$!>))

import Text.Parsec
    ( choice,
      many,
      try,
      ParsecT,
      )

import Ubc.Parse.Syntax.VariableType (VariableType)

import Ubc.Parse.Syntax.Function (Function)

import qualified Ubc.Parse.Syntax.Language as UbcLanguage
import qualified Ubc.Parse.Syntax.VariableType as VariableType
import qualified Ubc.Parse.Syntax.Function as Function

type VariableName = String

data StructStatement = Variable VariableName VariableType
        | Function Function.Function

data Struct = Struct 
        { name :: String
        , variables :: [(VariableName, VariableType)]
        , functions :: [Function]
        }
        deriving (Show)

addVariable :: Struct -> VariableName -> VariableType -> Struct
addVariable (Struct sn vs fs) n t = Struct sn ((n, t): vs) fs

addFunction :: Struct -> Function -> Struct
addFunction (Struct sn vs fs) f = Struct sn vs (f:fs)

parse :: Monad m => ParsecT String u m Struct
parse = do
        _ <- UbcLanguage.reserved "struct"
        structIdentifier <- UbcLanguage.identifier

        foldl accumulateStruct (Struct structIdentifier [] []) <$!> UbcLanguage.braces (many structMember)

accumulateStruct :: Struct -> StructStatement -> Struct
accumulateStruct s (Variable n t) = addVariable s n t
accumulateStruct s (Function f)   = addFunction s f

structMember :: Monad m => ParsecT String u m StructStatement
structMember = choice [ structVariableOrFunction ]

structVariableOrFunction :: Monad m => ParsecT String u m StructStatement
structVariableOrFunction = do
        (typeName, identifier) <- try $ do
                typeName <- UbcLanguage.typeName
                objectIdentifier <- UbcLanguage.identifier
                return (VariableType.fromString typeName, objectIdentifier)
        choice
                [ parseVariable typeName identifier
                , Function <$!> Function.parsePrefixed typeName identifier
                ] -- TODO: Functions on structs

parseVariable :: Monad m => VariableType -> String -> ParsecT String u m StructStatement
parseVariable variableType variableName = do
        _ <- UbcLanguage.semicolon
        return $ Variable variableName variableType
